{****************************************************************************
**                                                                         **
**              Copyright (c) 2004 microOLAP Technologies LLC,             **
**                      Halturin A.P., Naumov D.A.                         **
**                              Source File                                **
**                                                                         **
****************************************************************************}
unit HNWorkThread;

interface

uses
  Windows, SysUtils, HNThread, HNSyncList, Ethernet, WinSock;

//-----------------------------------------------------------------------------
// HN_PKT_COUNT - the number of the packets in the pool.
//                In other stream implementations this value can be made dynamic,
//                depending on the user defined options, the speed of the packet
//                handling algorithm and on the system resources 
//                (process speed, memory volume).
Const HN_PKT_COUNT = 50000;

type
//-----------------------------------------------------------------------------
// PKT_POOL_ITEM - the structure for the pool element, which contains 
//                 information about a single packet.
PKT_POOL_ITEM = record
   Link       : LIST_ENTRY;
   NumberID   : ULARGE_INTEGER;
   TimeStamp  : LARGE_INTEGER;
   PacketSize : Integer;
   Pkt        : array [0..1513] of BYTE;    // 1514 - Ethernet max. packet size
  end;
  PPKT_POOL_ITEM = ^PKT_POOL_ITEM;

//-----------------------------------------------------------------------------
  THNWorkThread = class(THNThread)
    private
      hSem          : THandle;                // busy packets semaphore
      WaitHandles   : array[0..1] of THandle; // wait handles array
      FreeList      : THNSyncList;            // list with free packets
      BusyList      : THNSyncList;            // list with busy packets
      pPacketPool   : Pointer;                // packets pool
    public
      constructor Create;
      destructor  Destroy; override;

    public
      function  Execute() : DWORD ; override;
      procedure ClearBusyList();
      procedure InsertPacket(pPacket : Pointer; PacketSize : Integer; var pNumberID : ULARGE_INTEGER; var pTimeStamp : LARGE_INTEGER);
      procedure DumpPacket(pItem : PPKT_POOL_ITEM);
      function  PacketToStr(Buf : PChar; pPacket : Pointer; PacketSize : Integer; var pNumberID : ULARGE_INTEGER; var pTimeStamp : LARGE_INTEGER) : Integer;

  end;

implementation

uses
  MainUnit;
//---------------------------------------------------------------------------
// Create
//---------------------------------------------------------------------------
constructor THNWorkThread.Create;
 var
   i          : DWORD;
   pPkt       : PPKT_POOL_ITEM;
 begin
   inherited Create;

   FreeList := THNSyncList.Create();
   BusyList := THNSyncList.Create();

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Reserve memory for the packet pool in one step.
   pPacketPool := VirtualAlloc(nil,SizeOf(PKT_POOL_ITEM)*HN_PKT_COUNT,MEM_COMMIT,PAGE_READWRITE);
   pPkt := PPKT_POOL_ITEM(pPacketPool);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Place all packets from the pool to the list of free packets.
   for i:=0 to HN_PKT_COUNT-1 do
    begin
      FreeList.InsertHead(@(pPkt^.Link));
      pPkt := PPKT_POOL_ITEM(DWORD(pPkt) + SizeOf(PKT_POOL_ITEM));
    end;

 end;

//---------------------------------------------------------------------------
// Destroy
//---------------------------------------------------------------------------
destructor THNWorkThread.Destroy;
 begin
   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Free the packet pool memory
   VirtualFree(pPacketPool,0,MEM_RELEASE);

   FreeList.Free();
   BusyList.Free();
   inherited Destroy;
 end;

//-----------------------------------------------------------------------------
// Execute - thread working function
//-----------------------------------------------------------------------------
function THNWorkThread.Execute() : DWORD;
 var
   pItem          : PPKT_POOL_ITEM;
   WaitObj        : DWORD;
   SaveLimitCount : DWORD;
 begin
   SaveLimitCount := LimitCount;

   HNConsole.printf('Work thread start'#13#10,[]);

   hSem := CreateSemaphore(nil,0,HN_PKT_COUNT,nil);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Sace the handle array for processing them within the WaitForMultipleObjects
   // function called in the main cycle 
   WaitHandles[0] := hStopEvent;
   WaitHandles[1] := hSem;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // The main cycle for handling the packets
   while(True) do
    begin
      //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      // Waiting for the "thread exit" event or the "new packet recieved" event
      WaitObj := WaitForMultipleObjects(2,@WaitHandles,False,INFINITE);
      if(WaitObj = WAIT_OBJECT_0) then
       begin 
         //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         // The thread exit event fired - leave the packet handling
         // cycle
         Break;
       end
      else if(WaitObj = WAIT_OBJECT_0 + 1) then
       begin
         //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         // The semaphor fired, i.e. the new unhandled packet appeared 
         // in the list of busy packets.

         //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         // If we use the packet number limit, then check the limit and exit if
         // the limit is reached.
         if(UseLimit) then
          begin
            if(LimitCount = 0) then Break;  // Exit thread
            Dec(LimitCount);
          end;

         //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         // Take the first unhandled packet 
         pItem := PPKT_POOL_ITEM(BusyList.RemoveHead());
         if(pItem <> nil) then
          begin
            //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            // Handle the new packet.
            // In our case we output it to the dump console and to the log file,
            DumpPacket(pItem);
            
            //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
            // Return the released packet to the list of free packets.
            // Add it to the end of the free packet list.
            FreeList.InsertTail(@(pItem^.Link));
          end;
       end;
    end;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Close the semaphore
   CloseHandle(hSem);

   if(not UseLimit) then
    begin
      HNConsole.printf('Work thread stopped.'#13#10,[]);
    end
   else
    begin
      HNConsole.printf('Work thread stopped: dump limit is set in %u.'#13#10,[SaveLimitCount]);
    end;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Send the message to the main window to inform it of the successful thread
   // exit.
   PostMessage(MainForm.Handle,HN_WM_THREAD_EXIT,0,0);

   Result := 0;
 end;

//-----------------------------------------------------------------------------
// ClearBusyList - clear the list of the busy packets.
//-----------------------------------------------------------------------------
procedure THNWorkThread.ClearBusyList();
 var
   pItem    : PPKT_POOL_ITEM;
 begin
   pItem := PPKT_POOL_ITEM(BusyList.RemoveHead());
   while(pItem <> nil) do
    begin
      FreeList.InsertTail(@pItem.Link);
      pItem := PPKT_POOL_ITEM(BusyList.RemoveHead());
    end;
 end;

//-----------------------------------------------------------------------------
// InsertPacket - add the new incoming packet to the queue for handling
//-----------------------------------------------------------------------------
procedure THNWorkThread.InsertPacket(pPacket : Pointer; PacketSize : Integer; var pNumberID : ULARGE_INTEGER; var pTimeStamp : LARGE_INTEGER);
 var
   pItem   : PPKT_POOL_ITEM;
 begin
   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Get the free packet from the list of free packets
   pItem := PPKT_POOL_ITEM(FreeList.RemoveHead());
   
   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // If there no free packets, then we lose this packet
   if(pItem = nil) then 
    begin
      Exit;
    end;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Fill the packet with incoming packet data
   pItem^.NumberID   := pNumberID;
   pItem^.TimeStamp  := pTimeStamp;
   pItem^.PacketSize := PacketSize;
   MoveMemory(@(pItem^.Pkt),pPacket,PacketSize);


   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Add the filled packet to the end of the list of packets, which require 
   // handling
   BusyList.InsertTail(@(pItem^.Link));

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Release the semaphore to inform the thread of the new packet requiring
   // handling in the list of busy packets
   ReleaseSemaphore(hSem,1,nil);
 end;

//-----------------------------------------------------------------------------
// DumpPacket - Output the packet dump to the console and to file 
//-----------------------------------------------------------------------------
procedure THNWorkThread.DumpPacket(pItem : PPKT_POOL_ITEM);
 var
   DumpBuff : array[0..(1024*10)] of Char;
   DumpLen  : Integer;
   bw       : DWORD;
 begin

   DumpBuff[0] := #0;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Save the dump as a string to buffer
   DumpLen := PacketToStr(DumpBuff,@pItem^.Pkt,pItem^.PacketSize,pItem^.NumberID,pItem^.TimeStamp);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Display the dump string on the console
   if(DumpToConsole) then HNConsole.WriteToConsole(DumpBuff,DumpLen);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Save the dump string to file 
   if(DumpToFile) then WriteFile(hLogFile,DumpBuff,DumpLen,bw,nil);
 end;

//---------------------------------------------------------------------------
// PacketToStr - Save the packet as a string to buffer and return the length
//               of the saved string
//---------------------------------------------------------------------------
function THNWorkThread.PacketToStr(Buf : PChar; pPacket : Pointer; PacketSize : Integer; var pNumberID : ULARGE_INTEGER; var pTimeStamp : LARGE_INTEGER) : Integer;
 var
   pPkt    : PByteArray;
   pEth    : PETHERNET_HEADER;
   EthType : WORD;
   SrcMAC  : PByteArray;
   DstMAC  : PByteArray;
   Str     : array[0..256] of Char;
   Ind     : Integer;
   i       : Integer;
   CurByte : Byte;
 begin
   pPkt    := PByteArray(pPacket);
   pEth    := PETHERNET_HEADER(pPacket);
   EthType := htons(pEth^.EthType);
   SrcMAC  := PByteArray(@(pEth^.EthSHost));
   DstMAC  := PByteArray(@(pEth^.EthDHost));

   StrCat(Buf,'=============================================================================='#13#10);
   StrCat(Buf,' Packet ID: ');
   StrCat(Buf,PChar(IntToStr(pNumberID.LowPart)));
   StrCat(Buf,'  Time: HighPart ');
   StrCat(Buf,PChar(IntToStr(pTimeStamp.HighPart)));
   StrCat(Buf,'  LowPart ');
   StrCat(Buf,PChar(IntToStr(pTimeStamp.LowPart)));
   StrCat(Buf,#13#10);
   StrCat(Buf,' Packet size: ');
   StrCat(Buf,PChar(IntToStr(PacketSize)));
   StrCat(Buf,' bytes '#13#10);
   StrCat(Buf,StrFmt(Str,' Eth type : 0x%4X'#13#10,[EthType]));
   StrCat(Buf,StrFmt(Str,' Src MAC  : %02X-%02X-%02X-%02X-%02X-%02X'#13#10,[SrcMAC[0],SrcMAC[1],SrcMAC[2],SrcMAC[3],SrcMAC[4],SrcMAC[5]]));
   StrCat(Buf,StrFmt(Str,' Dst MAC  : %02X-%02X-%02X-%02X-%02X-%02X'#13#10,[DstMAC[0],DstMAC[1],DstMAC[2],DstMAC[3],DstMAC[4],DstMAC[5]]));

   StrCat(Buf,'------------------------------ Packet dump -----------------------------------'#13#10);

   Ind := 0;

   while(Ind < PacketSize) do
    begin
      StrCat(Buf,PChar(IntToHex(Ind,4)));
      StrCat(Buf,':  ');
      for i:=0 to 15 do
       begin
         CurByte := pPkt[Ind + i];

         if(Ind + i < PacketSize) then StrCat(Buf,PChar(IntToHex(CurByte,2)))
                                  else StrCat(Buf,'  ');
         StrCat(Buf,' ');
         if(i = 7) then StrCat(Buf,'| ');
       end;

      StrCat(Buf,'    ');

      for i:=0 to 15 do
       begin
         CurByte := pPkt[Ind + i];
         Str[1]  := #0;
         if(Ind + i < PacketSize) then
          begin
            if(($19 < CurByte) and (CurByte < $7F)) then Str[0] := Char(CurByte)
                                                    else Str[0] := '.';
          end
         else Str[0] := ' ';

         StrCat(Buf,Str);
       end;

      StrCat(Buf,#13#10);
      Ind := Ind + 16;
    end;

   StrCat(Buf,#13#10);
   Result := StrLen(Buf);

 end;                                    

end.


