{****************************************************************************
**                                                                         **
**                             HNLogConsole                                **
**              Copyright (c) 2004 microOLAP Technologies LLC,             **
**                      Halturin A.P., Naumov D.A.                         **
**                              Source File                                **
**                                                                         **
****************************************************************************}

unit HNLogConsole;

interface
uses
  Windows, SysUtils;

type
  THNLogConsole = class
    private
      hOut     : THandle;
      XWinSize : Integer;   // X window size, ( <= 0 ) - for max size
      YWinSize : Integer;   // Y window size, ( <= 0 ) - for max size
      Title    : array[0..1024] of Char;

    public
      constructor Create;
      destructor  Destroy; override;

    public
      function  GetConsoleHwnd() : HWND;
      function  printf(Format: PChar; const Args: array of const) : Integer;
      function  WriteToConsole(pStr : PChar; StrLen : DWORD) : BOOL;
      function  SetTitle(pTitle : PChar) : BOOL;
      function  IsOpen() : BOOL;
      function  Close() : BOOL;
      procedure SetWinSize(XSize : Integer; YSize : Integer);
      function  Open(CmdShow : Integer) : BOOL;

  end;

implementation

function CtrlHandler(fdwCtrlType : DWORD) : BOOL;
 begin
   Result := True;
 end;

//---------------------------------------------------------------------------
// Create
//---------------------------------------------------------------------------
constructor THNLogConsole.Create;
 begin
   inherited Create;
   hOut     := INVALID_HANDLE_VALUE;
   XWinSize := 0;
   YWinSize := 0;
   Title    := '';
 end;

//---------------------------------------------------------------------------
// Destroy
//---------------------------------------------------------------------------
destructor THNLogConsole.Destroy;
 begin
   Close();
   inherited Destroy;
 end;

//---------------------------------------------------------------------------
// GetConsoleHwnd - get console window handle
//---------------------------------------------------------------------------
function THNLogConsole.GetConsoleHwnd() : HWND;
 var
   hwndFound         : HWND;                   // this is what is returned to the caller
   pszNewWindowTitle : array[0..1023] of Char; // contains fabricated WindowTitle
   pszOldWindowTitle : array[0..1023] of Char; // contains original WindowTitle
 begin

   // fetch current window title
   GetConsoleTitle(pszOldWindowTitle, 1024);

   // format a "unique" NewWindowTitle
   StrLFmt(pszNewWindowTitle,1024,'%d/%d',[GetTickCount(),GetCurrentProcessId()]);

   // change current window title
   SetConsoleTitle(pszNewWindowTitle);

   // ensure window title has been updated
   Sleep(70);

   // look for NewWindowTitle
   hwndFound := FindWindow(nil,pszNewWindowTitle);

   // restore original window title
   SetConsoleTitle(pszOldWindowTitle);

   Result := hwndFound;
 end;

//---------------------------------------------------------------------------
// printf - format output for short strings (up to 1000 bytes)
//---------------------------------------------------------------------------
function THNLogConsole.printf(Format: PChar; const Args: array of const) : Integer;
 var
   Buf : array[0..1028] of Char;
 begin
   StrLFmt(Buf,1028,Format,Args);
   Result := StrLen(Buf);
   WriteToConsole(Buf,Result);
 end;
   
//---------------------------------------------------------------------------
// WriteToConsole - console events handler
//---------------------------------------------------------------------------
function THNLogConsole.WriteToConsole(pStr : PChar; StrLen : DWORD) : BOOL;
 var
   BytesWritten : DWORD;
 begin
   if(hOut <> INVALID_HANDLE_VALUE) then Result := WriteFile(hOut,pStr^,StrLen,BytesWritten,nil)
                                    else Result := False;
 end;

//---------------------------------------------------------------------------
// IsOpen
//---------------------------------------------------------------------------
function THNLogConsole.IsOpen() : BOOL;
 begin
   Result := (hOut <> INVALID_HANDLE_VALUE);
 end;

//---------------------------------------------------------------------------
// SetTitle - set console title
//---------------------------------------------------------------------------
function THNLogConsole.SetTitle(pTitle : PChar) : BOOL;
 begin 
   StrCopy(Title,pTitle);
   if(IsOpen()) then SetConsoleTitle(Title);
   Result := True;
 end;

//---------------------------------------------------------------------------
// SetWinSize
//---------------------------------------------------------------------------
procedure THNLogConsole.SetWinSize(XSize : Integer; YSize : Integer);
 begin
   XWinSize := XSize;
   YWinSize := YSize;
 end;

//---------------------------------------------------------------------------
// Close - close console
//---------------------------------------------------------------------------
function THNLogConsole.Close() : BOOL;
 begin
   if(hOut <> INVALID_HANDLE_VALUE) then CloseHandle(hOut);
   hOut := INVALID_HANDLE_VALUE;
   FreeConsole();
   Result := True;
 end;

//---------------------------------------------------------------------------
// Open - open console
//---------------------------------------------------------------------------
function THNLogConsole.Open(CmdShow : Integer) : BOOL;
 var
   MaxSize        : COORD;
   CurSize        : COORD;
   ConsoleWinRect : SMALL_RECT;
   hCWnd          : HWND;
   hCMenu         : HMENU;
 begin

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // allocate console
   if(not AllocConsole()) then
    begin
      Result := False;
      Exit;
    end;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Delete "close" button from console window system menu
   hCWnd  := GetConsoleHwnd();
   hCMenu := GetSystemMenu(hCWnd,False);
   DeleteMenu(hCMenu, SC_CLOSE, MF_BYCOMMAND);

   SetConsoleTitle(Title);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Get out handle
   hOut := CreateConsoleScreenBuffer(GENERIC_READ or GENERIC_WRITE,0,nil,CONSOLE_TEXTMODE_BUFFER,nil);
   if(hOut = INVALID_HANDLE_VALUE) then
    begin
     Result := False;
     Exit;
    end;
   SetConsoleActiveScreenBuffer(hOut);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // set event handler
   SetConsoleCtrlHandler(@CtrlHandler,True);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // set console buffer size
   MaxSize := GetLargestConsoleWindowSize(hOut);

   if(XWinSize>0) then CurSize.X := XWinSize
   else
    begin
      if(YWinSize>MaxSize.Y) then CurSize.X := MaxSize.X - 3
                             else CurSize.X := MaxSize.X - 1 ;
    end;

   if(YWinSize>0) then CurSize.Y := YWinSize
                  else CurSize.Y := MaxSize.Y - 1;

   if(not SetConsoleScreenBufferSize(hOut,CurSize)) then
    begin
      Result := False;
      Exit;
    end;

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // set console window size
   ConsoleWinRect.Left   := 0;
   ConsoleWinRect.Top    := 0;
   ConsoleWinRect.Right  := CurSize.X - 1;
   if(ConsoleWinRect.Right > MaxSize.X) then ConsoleWinRect.Right := MaxSize.X;
   ConsoleWinRect.Bottom  := CurSize.Y - 1;
   if(ConsoleWinRect.Bottom > MaxSize.Y) then ConsoleWinRect.Bottom := MaxSize.Y;

   ConsoleWinRect.Right := 0;

   SetConsoleWindowInfo(hOut, True, ConsoleWinRect);

   //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
   // Show console window
   ShowWindow(hCWnd,CmdShow);

   Result := True;
 end;

end.


